
import numpy as np

from scipy.sparse import hstack, issparse, lil_matrix

import copy
from ..Base.ProblemTransformation import ProblemTransformationBase
from ..Base.base import MLSelectorBase

class Entropy_Label_Assignment(ProblemTransformationBase):
    
    def __init__(self, selector=None, require_dense=None):
        super(Entropy_Label_Assignment, self).__init__(selector, require_dense)
        self.selector = selector
    
    def calculate_entropy(self, probabilities):
        entropy = -np.sum(probabilities * np.log(probabilities + 1e-15))
        return entropy
    
    def g(self, entropy):
        return np.exp(-entropy)
    
    def ELA_label_assignment(self, X, y):
        n_samples, n_classes = y.shape
        transformed_labels = np.zeros(n_samples, dtype=int)
        
        for i in range(n_samples):
            class_probabilities = np.zeros(n_classes)
            for j in range(n_classes):
                entropy_for_class = self.calculate_entropy(y[i, j])
                class_probabilities[j] = self.g(entropy_for_class)
            class_probabilities /= np.sum(class_probabilities)
            transformed_labels[i] = np.argmax(class_probabilities)
        
        return transformed_labels
    
    def transform(self, y):
        transformed_labels = self.ELA_label_assignment(None, y)  # Pass None for X as it's not used
        return transformed_labels
    
    def fit(self, X, y):
        X = self._ensure_input_format(X, sparse_format="csr", enforce_sparse=True)

        self.selector.fit(self._ensure_input_format(X), self.transform(y))

        return self
     
    
    def fit_transform(self, X, y):
        selection = self.selector.fit_transform(X, self.transform(y))
        
        return selection
    
    def get_support(self, indices=True):
        indices = self.selector.get_support()
        return indices




from scipy import sparse

from ..Base.ProblemTransformation import ProblemTransformationBase

from ..Base.base import MLSelectorBase

import copy
import numpy as np
from scipy import sparse


class LabelPowerset(ProblemTransformationBase):
    
    
    def __init__(self, selector=None, require_dense=None):
        super(LabelPowerset, self).__init__(selector, require_dense)
        self.selector=selector
        
        
   
        self._clean()

    def _clean(self):
        """Reset classifier internals before refitting"""
        self.unique_combinations_ = {}
        self.reverse_combinations_ = []
        self._label_count = None
       
    def transform(self, y):
        """Transform multi-label output space to multi-class
        Transforms a mutli-label problem into a single-label multi-class
        problem where each label combination is a separate class.
        Parameters
        -----------
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        Returns
        -------
        numpy.ndarray of `{0, ... , n_classes-1}`, shape=(n_samples,)
            a multi-class output space vector
        """

        y = self._ensure_output_format(y, sparse_format="lil", enforce_sparse=True)

        self._clean()
        self._label_count = y.shape[1]

        last_id = 0
        train_vector = []
        for labels_applied in y.rows:
            label_string = ",".join(map(str, labels_applied))

            if label_string not in self.unique_combinations_:
                self.unique_combinations_[label_string] = last_id
                self.reverse_combinations_.append(labels_applied)
                last_id += 1

            train_vector.append(self.unique_combinations_[label_string])

        return np.array(train_vector)


    def fit(self, X, y):
        """Fits classifier to training data
        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        Returns
        -------
        self
            fitted instance of self
        Notes
        -----
        .. note :: Input matrices are converted to sparse format internally if a numpy representation is passed
        """
  
        X = self._ensure_input_format(X, sparse_format="csr", enforce_sparse=True)

        self.selector.fit(self._ensure_input_format(X), self.transform(y))

        return self

    def fit_transform(self, X,y):
        """Predict labels for X
        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        Returns
        -------
        :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
            
        """
  
        self.selector.fit(self._ensure_input_format(X), self.transform(y))

        # this will be an np.array of integers representing classes
        lp_prediction = self.selector.fit_transform(self._ensure_input_format(X), self.transform(y))

        return lp_prediction



    def inverse_transform(self, y):
        """Transforms multi-class assignment to multi-label
        Transforms a mutli-label problem into a single-label multi-class
        problem where each label combination is a separate class.
        Parameters
        -----------
        y : numpy.ndarray of `{0, ... , n_classes-1}`, shape=(n_samples,)
            binary indicator matrix with label assignments
        Returns
        -------
        :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        """
        n_samples = len(y)
        result = sparse.lil_matrix((n_samples, self._label_count), dtype="i8")
        for row in range(n_samples):
            assignment = y[row]
            result[row, self.reverse_combinations_[assignment]] = 1

        return result
    
    def get_support(self, indices=True):
        
        indices = self.selector.get_support()
        return indices


"""
The :mod:`Scikit-PTM-FS.base` module implements base for multi-label feature selection.

Two base classes are in use currently in Scikit-PTM-FS:

- :class:`MLSelectorBase` - a generic base class for multi-label feature selection
- :class:`ProblemTransformationBase` - the base class for problem transformation and ensemble approaches that handles a base selector
"""

from .base import MLSelectorBase
from .ProblemTransformation import ProblemTransformationBase

__all__ = ["MLSelectorBase", "ProblemTransformationBase"]



import numpy as np
import itertools

from scipy.sparse import hstack, issparse, lil_matrix
import copy

from ..Base.ProblemTransformation import ProblemTransformationBase
from ..Base.base import MLSelectorBase


class PairwiseComparsion(ProblemTransformationBase):
    
    def __init__(self, selector=None, require_dense=None):
        super(PairwiseComparsion, self).__init__()
        
        
        self.selector = selector
        self.require_dense = require_dense
        self.copyable_attrs = [
            "selector",
            "require_dense",
    
            
        ]
    def transform(self, X, Y):
        label_count = Y.shape[1]
        X_pairwise_list = []
        y_trans_list = []
        
        # Iterate through all pairs of labels (i, j) where i < j
        for label1, label2 in itertools.combinations(range(label_count), 2):
            
            # Determine indices of samples with label1 and label2
            indices_label1 = (Y[:, label1] == 1)
            indices_label2 = (Y[:, label2] == 1)
            
            # Select samples where labels are either label1 or label2, but not both
            mask = (indices_label1 | indices_label2) & ~(indices_label1 & indices_label2)
            X_pairwise = X[mask]

            # Combine label1 and label2 into a single label array
            y_pairwise = np.where(indices_label1[mask], label1, label2)
            X_pairwise_list.append(X_pairwise)
            y_trans_list.append(y_pairwise)
            
            
            # Find the maximum number of samples among all pairs
        max_samples = max(len(pair) for pair in X_pairwise_list)
        # Fill arrays to have the same number of samples (if needed)
        
        for i in range(len(X_pairwise_list)):
            pair_samples = len(X_pairwise_list[i])
            if pair_samples < max_samples:
                missing_samples = max_samples - pair_samples
                X_pairwise_list[i] = np.vstack((X_pairwise_list[i], X_pairwise_list[i][:missing_samples]))
                y_trans_list[i] = np.hstack((y_trans_list[i], y_trans_list[i][:missing_samples]))
        # Convert lists to arrays
        self.x_trans = np.vstack(X_pairwise_list)
        self.y_trans = np.hstack(y_trans_list)
        
        return self.x_trans, self.y_trans
    
    def fit(self, X, y):
        """Fits selector to training data
        Parameters
        ----------
        X : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix, shape=(n_samples, n_features)
            input feature matrix
        y : `array_like`, :class:`numpy.matrix` or :mod:`scipy.sparse` matrix of `{0, 1}`, shape=(n_samples, n_labels)
            binary indicator matrix with label assignments
        Returns
        -------
        self
            fitted instance of self
        Notes
        -----
        .. note :: Input matrices are converted to sparse format internally if a numpy representation is passed
        """
        x_trans, y_trans=self.transform(X,y)
     
        self.selector.fit(self._ensure_input_format(x_trans), self._ensure_output_format(y_trans))
        

        return self
        
     

    def fit_transform(self, X, y):
        """
        Fit the selector and return the transformed features.
        """
        x_trans, y_trans = self.transform(X, y)  # Ensure x_trans and y_trans are created
        selected_features = self.selector.fit_transform(
        self._ensure_input_format(x_trans),
        self._ensure_output_format(y_trans))
        return selected_features

    
    def get_support(self, indices=True):
        indices = self.selector.get_support()
        return indices
    




from setuptools import setup, find_packages
import os

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='scikit-ptm-fs',
    version='0.1.8',
    author='Omaima Al Hosni',
    author_email='alhosni_o@hotmail.com',
    description='A Python package for integrating scikit-learn feature selection with multi-label problem transformation methods.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Omaimah-AlHosni/scikit-ptm-fs',
    project_urls={
        'Source': 'https://github.com/Omaimah-AlHosni/scikit-ptm-fs',
        'Tracker': 'https://github.com/Omaimah-AlHosni/scikit-ptm-fs/issues',
    },
    packages=find_packages(include=['scikit_ptm_fs', 'scikit_ptm_fs.*']),
    install_requires=[
        'numpy>=1.20',
        'scikit-learn>=1.0',
        'scipy>=1.5',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    python_requires='>=3.7',
)


import copy
import numpy as np
from ..Base.ProblemTransformation import ProblemTransformationBase
from scipy.sparse import hstack, issparse, lil_matrix

class BinaryRelevance(ProblemTransformationBase):
    
    def __init__(self, selector=None, require_dense=None):
        super(BinaryRelevance, self).__init__(selector=selector, require_dense=require_dense)
        self.selector = selector

    def _generate_partition(self, X, y):
        """
        Partitions the label space into individual labels (singletons).
        """
        self.partition_ = list(range(y.shape[1]))
        self.model_count_ = y.shape[1]

    def fit(self, X, y):
        """
        Fits the selector independently for each label.
        """
        X = self._ensure_input_format(X, sparse_format="csr", enforce_sparse=True)
        y = self._ensure_output_format(y, sparse_format="csc", enforce_sparse=True)

        self.selectors_ = []
        self._generate_partition(X, y)
        self._label_count = y.shape[1]

        for i in range(self.model_count_):
            selector = copy.deepcopy(self.selector)
            y_subset = self._generate_data_subset(y, self.partition_[i], axis=1)
            if issparse(y_subset) and y_subset.shape[1] == 1:
                y_subset = np.ravel(y_subset.toarray())
            selector.fit(
                self._ensure_input_format(X),
                self._ensure_output_format(y_subset)
            )
            self.selectors_.append(selector)

        return self

    def fit_transform(self, X, y):
        """
        Fits and transforms the data for each label.
        Returns a list of transformed feature subsets (one per label).
        """
        self.fit(X, y)

        selected_features_per_label = []

        for i in range(self.model_count_):
            transformed = self.selectors_[i].transform(self._ensure_input_format(X))
            selected_features_per_label.append(transformed)

        return selected_features_per_label

    def get_support(self, indices=True, merge=False):
        """
        Returns the selected feature indices or masks per label.
        
        Parameters
        ----------
        indices : bool
            If True, returns indices; otherwise, returns a boolean mask.
        merge : bool
            If True, merges the selected supports across all labels.

        Returns
        -------
        list of supports per label or a merged support (union).
        """
        support_all = []
        for label in range(self.model_count_):
            support = self.selectors_[label].get_support(indices=indices)
            support_all.append(support)

        if merge:
            if indices:
                # Merge indices by union and return sorted list
                return sorted(set(np.concatenate(support_all)))
            else:
                # Merge masks using logical OR
                return np.any(np.vstack(support_all), axis=0)

        return support_all

  